# -*- coding: utf-8 -*-
r"""
This module contains utilities for the visualization of spin configurations:
- calculation of the color code
- theta, phi, length representation
"""
import numpy as np
from typing import Tuple, List


def identifylayers(position: np.ndarray) -> Tuple[int, np.ndarray]:
    r"""
    Calculates the number of layers and the z-coordinates of these layers.

    Args:
        position: position of the spins or eigenvectors (or any other given quantity written in STM-format).
        Format [[x1,y1,z1], ..., [xn,yn,zn]]

    Returns:
        the number of layers (int) and the z-coordinates of these layers (np.ndarray). The z-coordinates are ordered in
        an increasing order, due to the use of numpys unique function
    """
    l_z = np.unique(position[:, 2])
    l_n = len(l_z)
    return l_n, l_z


def determine_layer_centers(position: np.ndarray) -> List[np.ndarray]:
    r"""
    Takes the positions of the 3d lattice and first split up the layers. Then determine the center of each layer as
    the centroid.

    Args:
        position(np.ndarray): positions of the objects. Format [[x1,y1,z1],...,[xn,yn,zn]]

    Returns:
        center for each layer [[cx1,cy1,cz1],...,[cxL,cyL,czL]], where L is the number of layers.
    """
    l_z = np.unique(position[:, 2])
    centers = []
    for zz in l_z:
        centers.append(np.mean(position[position[:, 2] - zz <= 1e-3], axis=0))
    return centers


def tabledist(position: np.ndarray, referencepoint: np.ndarray = np.array([0.0, 0.0])) -> np.ndarray:
    r"""
    Calculates the distances to the neighbor shells of a set of 2D positions with reference point at (0, 0).

    Args:
        position(np.ndarray): Positions of spins or vectors( or any other given quantity written in the STM-format)
        within the xy-plane

        referencepoint(np.ndarray): Reference point in the plane for the distance calculations

    Returns:
        distances to neighbor shells from reference point within 2d plane
    """
    if np.shape(position)[1] != 2 or np.shape(referencepoint)[0] != 2:
        raise ValueError('This tabledist-function only accepts positions within a plane with two coordinates.')
    return np.unique(np.round_(np.linalg.norm(position - referencepoint, axis=1), decimals=5))


def calculate_phi_theta(spin: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    r"""
    Calculates the spherical coordinates of an array of spins or vectors:

    Args:
        spin(np.ndarray): spins of the spin configuration (could be used for eigenvectors as well).
        Format [[sx1,sy1,sz1],...,[sxn,syn.szn]]

    Returns:
        phi(np.ndarray): polar angle for each spin in degree
        theta(np.ndarray): azimutal angle for each spin in degree
        length(np.ndarray): length of spin. Should be 1 for spins but different from one in the case of eigenvectors.
    """
    l_length = np.linalg.norm(spin, axis=1)
    l_length_ip = np.linalg.norm(spin[:, :2], axis=1)
    l_phi = np.arctan2(spin[:, 1], spin[:, 0]) * 180 / np.pi
    l_theta = np.arctan2(l_length_ip, spin[:, 2]) * 180 / np.pi
    return l_phi, l_theta, l_length


def calculate_colorcode_azimut(theta: np.ndarray) -> List[List[float]]:
    r"""
    Calculates the color code (adapted from SpinD code) for the azimutal orientation of spins. Spins pointing into the
    +z-direction will be painted in red as spins pointing in the -z-direction will be painted in blue. In plane spins
    will be painted in green no matter the polar angle.

    Args:
        phi(np.ndarray): polar angle for each spin in degree
        theta(np.ndarray): azimutal angle for each spin in degree

    Returns:
        the colorcode for the azimutal angle for each spin [[R1,B1,G1],...,[Rn,Bn,Gn]]
    """
    colorcode = []
    for th in theta:
        C = 2 * np.pi / 3
        D = 5
        E = np.pi * 2 * th / 300
        R = D * np.cos(E)
        G = D * np.cos(E + C)
        B = D * np.cos(E + 2 * C)
        if R <= 1e-6:
            R = 0
        if G <= 1e-6:
            G = 0
        if B <= 1e-6:
            B = 0
        colorcode.append([R, B, G])
    return colorcode


def angle_between(v1, v2):
    """
    Returns:
         the angle in radians between vectors 'v1' and 'v2':
    """
    v1_u = v1 / np.linalg.norm(v1)
    v2_u = v2 / np.linalg.norm(v2)
    return np.math.atan2(np.linalg.det([v1,v2]),np.dot(v1,v2))


def pointinpolygon(point: np.ndarray, vertices: np.ndarray) -> bool:
    r"""
    If only a subset of the lattice shall be shown one has to calculate whether a given point lies within a polygon.
    This refers to the following problem and algorithms:
    https://en.wikipedia.org/wiki/Point_in_polygon
    This problem is normally adressed either through ray tracing or calculating the winding number. However both
    complexities are O(n), but the winding number calculation involves inverse trigonometric function that are comp.
    costly.

    Here a simple winding number approach is chosen.

    Args:
         point(np.ndarray): point which shall be tested to lie within a polygon

         vertices(np.ndarray): vertices of the polygon (they have to be in the order along the contour. (It doesn't
         matter if they're given cw or ccw). Format vertices = [[v1x,v1y,v1z],..., [vNx,vNy,vNz]], where N is the number
         of vertices.
    """
    l_anglesum = 0
    for (idx,vertex) in enumerate(vertices):
        if idx == len(vertices)-1:
            l_next_vertex = vertices[0]
        else:
            l_next_vertex = vertices[idx + 1]
        l_anglesum += angle_between(vertex - point, l_next_vertex - point)
    if abs(abs(l_anglesum) - 2 * np.pi) <= 1e-2:
        return True
    elif abs(l_anglesum) <= 1e-2:
        return False
    else:
        raise ValueError('Error during calculation of winding number')




# -*- coding: utf-8 -*-
r"""
Abstract class for the calculation of a volume of a goldstone mode. Possible examples are rotational modes for Skyrmions
or Antiskyrmions without DMI or the skyrmion or saddlepoint translation modes.
"""
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Union
from python3.magnetisations import MultiLayer

class IVolume(ABC):
    r"""
    abstract base class for goldstone mode volume calculations for a given STMfile.
    """

    @property
    @abstractmethod
    def STMfile(self) -> Path:
        r"""
        Returns:
             the path of the STM-file
        """

    @abstractmethod
    def __call__(self) -> List[float]:
        r"""
        Returns:
            the goldstone mode volume for all layers in the spin configuration in the STMfile with increasing z-
            component
        """

    @property
    @abstractmethod
    def ML(self) -> MultiLayer:
        r"""
        Returns:
            The magnetization read from the STM file. If the system is a monolayer the SL is accesible through the
            ML.layer[0] attribute.
        """



# -*- coding: utf-8 -*-
r"""
Simple test script for the sake of testing correct vapory installation. Vapory needs povray to be installed. If you're
on windows please make sure that the path to the povray executable is set to ...\pvengine.exe inside vapory.
"""
from vapory import *


def render_cube():
    r"""
    testing rendering of simple cube
    """
    scene = Scene(Camera('location', [0.0, 0.5, -4.0],
                         'direction', [0, 0, 1.5],
                         'look_at', [0, 0, 0]),
                  objects=[

                      Background("color", [0.85, 0.75, 0.75]),

                      LightSource([0, 0, 0],
                                  'color', [1, 1, 1],
                                  'translate', [-5, 5, -5]),

                      LightSource([0, 0, 0],
                                  'color', [0.25, 0.25, 0.25],
                                  'translate', [6, -6, -6]),

                      Box([-0.5, -0.5, -0.5], [0.5, 0.5, 0.5],
                          Texture(Pigment('color', [1, 0, 0]),
                                  Finish('specular', 0.6),
                                  Normal('agate', 0.25, 'scale', 0.5)),
                          'rotate', [45, 46, 47])
                  ]
                  )
    # We use antialiasing. Remove this option for faster rendering.
    scene.render("cube.png", width=300, height=300, antialiasing=0.001)



# -*- coding: utf-8 -*-
r"""
This module contains utilities for dealing with colors.
"""
import numpy as np


def RGB_255to1(colorcode: np.ndarray) -> np.ndarray:
    r"""
    converts RGB colorcodes from the 255 range within the range of povray which is 0 to 1.

    Args:
        colorcode(np.ndarray): [R, G, B] from 0 to 255

    Returns:
        colorcode [R, G, B] from 0 to 1
    """
    return colorcode / 255


def lightgray() -> np.ndarray:
    r"""
    Returns:
        lightgray colorcode
    """
    return RGB_255to1(np.array([207, 207, 207]))



# -*- coding: utf-8 -*-
r"""
Library for shell commands and sub processes.
Adapted from Moritz Goerzen's library 08.2018 (Spintronic Theory Kiel).
"""
import os
import platform
import subprocess
import sys
import shellcommands.constants as const
from contextlib import contextmanager
from pathlib import Path
from typing import Union, TypeVar


PathLike = TypeVar("PathLike", str, Path)

# operating system
OPERATINGSYSTEM = platform.system()


@contextmanager
def change_directory(newdir: PathLike) -> None:
    r"""
    Changes the directory using context manager. Will automatically put everything before the yield statement into the
    __enter__-method and everythin behind in the __exit__method.

    Args:
        newdir(str): name of the directory. Has to be a subdirectory of the actual directory.
    """
    prevdir = Path.cwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)


def copy_element(target: PathLike, destination: PathLike) -> None:
    r"""
    Copies an element (target) to a destination. Chooses the command based on OS. On windows the shell=True flag
    must be set. Should be replaced someday by shutil.which
    (https://stackoverflow.com/questions/3022013/windows-cant-find-the-file-on-subprocess-call)

    Args:
        target(PathLike): target file
        destination(PathLike): destination file
    """
    if OPERATINGSYSTEM == 'Windows':
        command = "copy " + str(target) + " " + str(destination)
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE, shell=True)
    elif OPERATINGSYSTEM == 'Linux':
        command = "cp " + str(target) + " " + str(destination)
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE)
    else:
        raise Exception('operating system not yet coded.')
    output, error = process.communicate()


def make_directory(directory: PathLike) -> None:
    r"""
    Creates directory. Chooses the command based on OS. It is possible to create multiple nested directories. It is
    highly recommended to use the input type of pathlibs Path then. Otherwise one has to set the correct back- or
    forward slashes. This destroys the benefit of this platform indepent implementation.

    Args:
        directory(PathLike): new directory
    """
    if OPERATINGSYSTEM == 'Windows':
        command = "md " + str(directory)
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE, shell=True)
    elif OPERATINGSYSTEM == 'Linux':
        command = "mkdir " + str(directory)
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE)
    else:
        raise Exception('operating system not yet coded.')
    output, error = process.communicate()


def copy_folder(target: PathLike, destination: PathLike) -> None:
    r"""
    Copies folder from target to destination. Chooses the command based on Os. For windows the /i flag supresses asking
    whether copying and the /e flag enables empty copying.
    """
    if OPERATINGSYSTEM == 'Windows':
        command = "xcopy " + str(target) + " " + str(destination) + "/i /e"
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE, shell=True)
    elif OPERATINGSYSTEM == 'Linux':
        command = "cp -r " + str(target) + " " + str(destination)
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE)
    else:
        raise Exception('operating system not yet coded.')
    output, error = process.communicate()


def remove_element(filename: PathLike) -> None:
    r"""
    Removes element (filename). Chooses the command based on Os. The windows /q flag is for silent removing. If the file
    is given as a path it is highly recomm. to use pathlibs Path for platform indepent input.
    """
    if OPERATINGSYSTEM == 'Windows':
        command = "del " + str(filename) + " /q"
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE, shell=True)
    elif OPERATINGSYSTEM == 'Linux':
        command = "rm" + str(filename)
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE)
    else:
        raise Exception('operating system not yet coded.')
    output, error = process.communicate()


def remove_folder(directory: PathLike) -> None:
    r"""
    Removes directory with everything inside. Chooses the command based on Os. The windows /s flag is for
    deleting subfolders and files and the /q flag is for silent modus.
    """
    if OPERATINGSYSTEM == 'Windows':
        command = "rmdir " + str(directory) + " /s /q"
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE, shell=True)
    elif OPERATINGSYSTEM == 'Linux':
        command = "rm -r" + str(directory)
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE)
    else:
        raise Exception('operating system not yet coded.')
    output, error = process.communicate()


def call_povray(script: PathLike, w: int = 2000, h: int = 2000) -> None:
    r"""
    Calls a povray script from the command line. For windows the environment variable pvengine has to be set. This
    interface is programmed for Povray3.7 GUI for windows.

    Args:
        script(PathLike): .pov file
        w(int, Optional): width. Default is 2000
        h(int, Optional): height. Default is 2000
    """
    if OPERATINGSYSTEM == 'Windows':
        command = 'pvengine /NR /EXIT /RENDER ' + str(script) + ' -d' + ' -w' + str(w) + ' -h' + str(h)
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE, shell=True)
    elif OPERATINGSYSTEM == 'Linux':
        command = "povray -d" + " -w" + str(w) + " -h" + str(h) + " " + str(script)
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE)
    else:
        raise Exception('operating system not yet coded.')
    output, error = process.communicate()


def call_spin(pathtoexecutable: PathLike = None) -> None:
    r"""
    Calls the spin D algorithm

    Args:
        pathtoexecutable(PathLike, Optional): the path to the SpinD executable. The default is None. In that case the
        default Linux or Windows paths is chosen based on the OS.
    """
    if pathtoexecutable is None:
        if OPERATINGSYSTEM == 'Windows':
            command = const.SPIND_WINDOWS
        elif OPERATINGSYSTEM == 'Linux':
            command = const.SPIND_LINUX
        else:
            raise Exception('operating system not yet coded.')
    else:
        command = pathtoexecutable
    with open('job.out','a') as f:
        process = subprocess.Popen(command.split(), stdout=f, shell=True)
    output, error = process.communicate()



# -*- coding: utf-8 -*-
r"""
Classes for filtering the spin or eigenvector lattices that are load from source. Contains a base abstract filter
class and inheriting ones.
"""
from abc import ABC, abstractmethod
from povray.utilities.spinconfiguration_utilities import pointinpolygon
import numpy as np


class CFilterLattice(ABC):
    r"""
    Abstract base class for filtering lattices
    """

    @abstractmethod
    def applyfilter(self, tofilter: np.ndarray) -> np.ndarray:
        r"""
        """

    @abstractmethod
    def _filter(self, performfilteron: np.ndarray) -> np.ndarray:
        r"""
        """

    @abstractmethod
    def __init__(self, performfilteron: np.ndarray) -> None:
        r"""
        Initializes the filtering.

        Args:
            performfilteron(np.ndarray): The quantity to which the filter condition applies. Every quantity with the
            same shape as this quantity can be filtered based on the filtering performed for this input array using the
            method applyfilter.
        """


class CFilterHexAroundCenter(CFilterLattice):
    r"""
    Filters a hexagonal lattice around the center of the lattice.
    """

    def applyfilter(self, tofilter: np.ndarray) -> np.ndarray:
        r"""
        Performs the filtering based on the boolean condition derived for the quantity which was initialized in constr.
        the filter.

        Args:
            tofilter(np.ndarray): quantity which shall be filtered based one this filter instance. Has to have the same
            shape as performfilteron

        Returns:
            the filtered quantity
        """
        ind = np.where(self._condition)
        return tofilter[ind]

    def _filter(self, performfilteron: np.ndarray) -> np.ndarray:
        r"""
        performs filtering. The filtering is performed in an hexagonal area around the center of the lattice.

        Args:
            performfilteron(np.ndarray): quantity on which the filter is derived.

        Returns:
            A boolean numpy array which can be used via np.extract(condition, otherarray)
        """
        l_centroid = np.mean(performfilteron, axis=0)
        l_edgelength = np.max(performfilteron[:, 0]) - np.min(performfilteron[:, 0])
        l_edge_reduced = l_edgelength * self._percentage
        l_polygon = np.array([[0.0, 0.0], [0.5 * l_edge_reduced, np.sqrt(3) / 2 * l_edge_reduced], [l_edge_reduced, 0],
                              [0.5 * l_edge_reduced, -1 * np.sqrt(3) / 2 * l_edge_reduced]]) + np.array(
            [l_edgelength / 2 * (1 - self._percentage), 0.0]) + self._polygonshift
        l_cond = []
        for position in performfilteron:
            l_cond.append(pointinpolygon(point=position[:2], vertices=l_polygon))
        return np.array(l_cond)

    def __init__(self, performfilteron: np.ndarray,polygonshift = np.array([0.0, 0.0]), percentage: float = 0.5) -> None:
        r"""
        initializes the filter

        Args:
             performfilteron(np.ndarray): quantity on which the filter is derived.
             percentage(float): percentage for edgelength_filtered / edgelength_unfiltered.
             polygonshift(np.ndarray): whether the polygon is shifted. This is necessary if the center of the filtering
             shouldn't be in the center of the lattice
        """
        super().__init__(performfilteron)
        self._percentage = percentage
        self._polygonshift = polygonshift
        self._condition = self._filter(performfilteron=performfilteron)



# -*- coding: utf-8 -*-
r"""
Class responsible for calculating in plane translational volumes of spin lattices.
"""
from typing import List

from python3.magnetisations import MultiLayer
from python3.propertyfunctions import zero_mode_partitionfunction
from STMutilities.goldstone.ivolume import IVolume
from pathlib import Path


class CTranslationVol(IVolume):
    r"""
    Class responsible for calculating in plane translational volumes of spin lattices. The inner methods used are taken
    from the userlib-spinD
    """

    def __init__(self, STMfile: Path) -> None:
        r"""
        Initializes the calculation of the translation volume

        Args:
            STMfile(Path): path to file with STM spin configuration
        """
        self._STMfile = STMfile
        self._ML = MultiLayer(path=str(self.STMfile))

    @property
    def STMfile(self) -> Path:
        r"""
        Returns:
             the path of the STM-file
        """
        return self._STMfile

    def __call__(self) -> List[float]:
        r"""
        Calls the zero_mode_partitionfunction from the userlib-spinD

        Returns:
            the translation mode volumes for each layer
        """
        vols = []
        for layer in self._ML.Layer:
            vols.append(zero_mode_partitionfunction(SL=layer,mode='translation'))
        return vols
    
    @property
    def ML(self) -> MultiLayer:
        r"""
        Returns:
            The magnetization read from the STM file. If the system is a monolayer the SL is accesible through the
            ML.layer[0] attribute.
        """
        return self._ML



# -*- coding: utf-8 -*-
r"""
Module containing different camera classes
"""
from vapory import Camera
from typing import Union, List, Tuple
from povray.predefined_povrayobjects.icamera import ICamera
import numpy as np


def rotationmatrix(angle: float) -> np.ndarray:
    r"""
    Calculates the rotationmatrix as need for the definition of the camera in povray (left handed system)

    Args:
        angle(float): angle of rotation around x axis

    Returns:
        rotation matrix as np ndarray
    """
    return np.array([[1, 0, 0],
                     [0, np.cos(np.radians(90 - angle)), -1 * np.sin(np.radians(90 - angle))],
                     [0, np.sin(np.radians(90 - angle)), np.cos(np.radians(90 - angle))]])


class CCameraForSpinLattice(ICamera):
    r"""
    Camera class for spin lattices. Might be used for spin configuration as well as for eigenvectors.
    """

    def __init__(self, position: np.ndarray, location: Union[None, List[float]] = None,
                 lookat: Union[None, List[float]] = None, viewangle: float = 90,
                 viewdistance: Union[float, None] = None, translatelookatz: float = 0.0) -> None:
        r"""
        initializes the camera which is accesible through the property cam.

        Args:
            position(np.ndarray): Numpy arrays of the positions within the spin lattice. Format [[x1,y1,z1],....]

            location(List[float]): location of the observer. Default value is None. In that case the location will be
            automatically set with the use of the positions of the spins in the source file.

            lookat(List[float]): location of the observed point. Default value is None. In that case the location will
            be automatically set with the use of the positions of the spins in the source file.

            viewangle(float): Angle for vector from location to lookat. Default angle is 90 degree, which corresponds
            to a view from the top of the spin configuration. Will be used only if location and lookat are None

            viewdistance(float): Distance from the camera to the observed point. Default distance is None. Will be used
            only if location and lookat are None and if it is different from None. In case that location and lookat are
            not defined, a viewangle is selected and this parameter is None the distance will be calculated automatic
            from the spin configuration itself.

            translatelookatz(float): shift observed point in z direction (used in multilayer systems)
        """
        self._camera = self._load_camera(position, location, lookat, viewangle, viewdistance, translatelookatz)

    @classmethod
    def _load_camera(cls, position: np.ndarray, location: Union[None, List[float]], lookat: Union[None, List[float]],
                     viewangle: float, viewdistance: Union[float, None], translatelookatz: float) -> Camera:
        r"""
        Loads the camera based either on user input or, per default, from spin configuration itself.

        Args:
            location(List[float]): location of observer
            lookat(List[float]): location of observed point
            viewangle(float): incident camera angle (90 means perpendicular to plane)
            viewdistance(float): distance of camera to observer
            translatelookatz(float): shift observed point in z direction (used in multilayer systems)

        Returns:
            An instance of vaporys (povrays) Camera
        """
        l_angle = 40
        if location is None and lookat is None:
            # determine center of spin configuration (in cartesians this the centroid of a point cloud is just the mean)
            # if any coordinate of the centroid is less than 1e-8 it is set to zero.
            l_lookat = np.array([0 if abs(coord) <= 1e-8 else coord for coord in (np.mean(position, axis=0))]) + np.array(
                [0.0, 0.0, translatelookatz])
            # rotate orthogonal vec
            l_lookdir = rotationmatrix(viewangle).dot(np.array([0, 0, 1]))
            # norm view direction and remove small entries
            l_lookdir_vec = [0 if abs(ld) <= 1e-8 else ld for ld in l_lookdir / np.linalg.norm(l_lookdir)]
            if viewdistance is None:
                # if viewdistance is not given calculate from extent of spin configuration in x direction
                extent_x = 0.5 * (np.max(position[:, 0]) - np.min(position[:, 0]))
                viewdistance = extent_x / np.tan(np.radians(l_angle / 2))
            l_location = l_lookat + viewdistance * np.array(l_lookdir_vec)
        elif location is None:
            l_lookat = lookat
            # rotate orthogonal vec
            l_lookdir = rotationmatrix(viewangle).dot(np.array([0, 0, 1]))
            # norm view direction and remove small entries
            l_lookdir_vec = [0 if abs(ld) <= 1e-8 else ld for ld in l_lookdir / np.linalg.norm(l_lookdir)]
            if viewdistance is None:
                # if viewdistance is not given calculate from extent of spin configuratin in x direction
                extent_x = 0.5 * (np.max(position[:, 0]) - np.min(position[:, 0]))
                viewdistance = extent_x / np.tan(np.radians(l_angle / 2))
            l_location = np.array(l_lookat) + viewdistance * np.array(l_lookdir_vec)
        elif lookat is None:
            l_location = location
            l_lookat = np.array([0 if abs(coord) <= 1e-8 else coord for coord in (np.mean(position, axis=0))]) + np.array(
                [0.0, 0.0, translatelookatz])
        else:
            l_location = location
            l_lookat = lookat

        return Camera('location', list(l_location), 'sky', [0, 0, 1], 'look_at', list(l_lookat), 'right', [1, 0, 0],
                      'angle', l_angle)

    @property
    def cam(self) -> Camera:
        r"""
        Returns:
            the camera
        """
        return self._camera



# -*- coding: utf-8 -*-
r"""
Abstract base class for all povray visualization that are created from a source file. E.g. SpinSTM-files or
Eigenvectors.
"""
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TypeVar

PathLike = TypeVar("PathLike", str, Path)


class IPovFromSource(ABC):
    r"""
    Abstract base class for visualization of source files
    """

    @abstractmethod
    def __repr__(self) -> str:
        r"""
        Returns:
            string representation of visualization class
        """

    @abstractmethod
    def __call__(self, outputpath: PathLike, width: float = 2000, height: float = 2000, antialiasing: bool = False,
                 antialiasingvalue: float = 0.001) -> None:
        r"""
        calls the rendering

        Args:
            outputpath(PathLike): name and path of created output
            width(float): width of the rendered image
            height(float): height of the rendered image
            antialiasing(bool): bool whether rendering is done with antialiasing
            antialiasingvalue(float): value for antialiasing
        """



# -*- coding: utf-8 -*-
r"""
Gets and summarizes the results of a short gneb calculation. Normally those calc. are not converged. They can be used
to identify intermediate minima in the path.
"""
import pandas as pd
import numpy as np
from results.igetresults import IResults
from pathlib import Path
from typing import Union, List, Tuple, Dict
from python3.magnetisations import SpinLattice, MultiLayer
from python3.shell_commands import *
from scipy.signal import argrelextrema
import os


class CGNShortGneb(IResults):
    r"""

    """

    def __init__(self, calc_dir: Path = Path.cwd(), nlayer: int = 1, signaturename: str = 'signature',
                 signaturevalue: Union[float, str] = 0, remove_energy_density: bool = False) -> None:
        r"""
        Initializes the result-summarizing process.

        Args:
            calc_dir(Path): directory of the spinD calculation. Default is the current working directory.

            nlayer(int): number of layers for the system.

            signaturename(str): if multiple calculations are evaluated with the goal to connect the dataframes then you
            can define a signaturename which the column names which describes what quantity changes from calc. to calc.
            It can be for example magnetic field or interlayer exchange

            signaturevalue(float, str): The value for the current calculation

            remove_energy_density(bool): remove energy density files after summarizing them. The default is False.
        """
        self.calc_dir = calc_dir
        self.nlayer = nlayer
        self.signaturename = signaturename
        self.signaturevalue = signaturevalue
        self.removeenergydensity = remove_energy_density
        self.i_en_path = (calc_dir / 'en_path.out').is_file()
        self.i_en_dens = (calc_dir / 'energy_density_end.dat').is_file()

    def __call__(self) -> pd.DataFrame:
        r"""
        """
        d = {self.signaturename: self.signaturevalue}
        d.update(self.enpathout())
        d.update(self.endensity())
        return pd.DataFrame([d], index=[self.signaturevalue])

    def endensity(self):
        r"""
        Reads the energy_density_end.dat file if the energy density file exists

        Returns:
            a dictionary containing the sum of the energy contributions for each layer for each image of the path
        """
        if not self.i_en_dens:
            endict = {}
            for idx in range(self.nlayer):
                endict[f'ex_intra_lay{idx}'] = None
                endict[f'ex_inter_lay{idx}'] = None
                endict[f'dmi_lay{idx}'] = None
                endict[f'ani_lay{idx}'] = None
                endict[f'zee_lay{idx}'] = None
                endict[f'dip_lay{idx}'] = None
                endict[f'biq_lay{idx}'] = None
                endict[f'four_lay{idx}'] = None
                endict[f'tot_lay{idx}'] = None
            return endict
        df = pd.read_csv(self.calc_dir / 'energy_density_end.dat', sep=r'\s+', index_col=False)
        df.columns = ['x', 'y', 'z', 'ex_intra', 'ex_inter', 'dmi', 'ani', 'zee', 'dip', 'biq', 'four', 'tot']
        z_layer = df.z.unique()

        imnr = []
        for f in os.listdir(str(self.calc_dir)):
            if str(f).startswith('energy_density_end'):
                pass
            elif str(f).startswith('energy_density_'):
                imnr.append(int(str(f)[15:-4]))

        endict = {}
        for (idx, zl) in enumerate(z_layer):
            ex_intra, ex_inter, dmi, ani, zee, dip, biq, four, tot = [], [], [], [], [], [], [], [], []
            for nr in imnr:
                df = pd.read_csv(self.calc_dir / f'energy_density_{nr}.dat', sep=r'\s+', index_col=False)
                df.columns = ['x', 'y', 'z', 'ex_intra', 'ex_inter', 'dmi', 'ani', 'zee', 'dip', 'biq', 'four', 'tot']
                dfl = df[df['z'] == zl]
                ex_intra.append(dfl.ex_intra.sum())
                ex_inter.append(dfl.ex_inter.sum())
                dmi.append(dfl.dmi.sum())
                ani.append(dfl.ani.sum())
                zee.append(dfl.zee.sum())
                dip.append(dfl.dip.sum())
                biq.append(dfl.biq.sum())
                four.append(dfl.four.sum())
                tot.append(dfl.tot.sum())

            endict[f'ex_intra_lay{idx}'] = np.array(ex_intra)
            endict[f'ex_inter_lay{idx}'] = np.array(ex_inter)
            endict[f'dmi_lay{idx}'] = np.array(dmi)
            endict[f'ani_lay{idx}'] = np.array(ani)
            endict[f'zee_lay{idx}'] = np.array(zee)
            endict[f'dip_lay{idx}'] = np.array(dip)
            endict[f'biq_lay{idx}'] = np.array(biq)
            endict[f'four_lay{idx}'] = np.array(four)
            endict[f'tot_lay{idx}'] = np.array(tot)

        if self.removeenergydensity:
            remove_element(self.calc_dir / 'energy_density_end.dat')
            for nr in imnr:
                remove_element(self.calc_dir / f'energy_density_{nr}.dat')

        return endict

    def enpathout(self) -> Dict[str, Union[np.ndarray, int, None]]:
        r"""
        Inspects en_path.out. There are three quantities of interest. We can check the total energy for intermediate
        minima.

        Returns:
            A nested dict: R, E, dE and the index of the intermediate minimum if it exists (otherwise None). Note that
            the image number of the intermed. min. is index + 1
        """
        if not self.i_en_path:
            return {'R': None, 'E': None, 'dE': None, 'min_idx': None}
        df = pd.read_csv(self.calc_dir / 'en_path.out', sep=r'\s+', usecols=[0, 1, 2], names=['R', 'E', 'dE'],
                         index_col=False)
        intermediate_imagenr = argrelextrema(df['E'].to_numpy(), np.less)[0]
        if intermediate_imagenr.size == 0:
            intermediate_imagenr = None
        return {'R': df['R'].to_numpy(), 'E': df['E'].to_numpy(), 'dE': df['dE'].to_numpy(),
                'min_idx': intermediate_imagenr}



# -*- coding: utf-8 -*-
r"""
Abstract class for getting results
"""
from abc import ABC, abstractmethod
import pandas as pd


class IResults(ABC):
    r"""
    abstract class for getting results of spinD calculations
    """

    @abstractmethod
    def __call__(self) -> pd.DataFrame:
        r"""
        calls the result-getting process
        """



# -*- coding: utf-8 -*-
r"""
abstract class for all cameras
"""
from vapory import Camera
from abc import ABC, abstractmethod


class ICamera(ABC):
    r"""
    abstract class for cameras
    """

    @property
    @abstractmethod
    def cam(self) -> Camera:
        r"""
        Returns:
             the camera
        """


# -*- coding: utf-8 -*-
r"""
Classes for visualization of spin configurations from STM files.

-CPovSTMSimpleMono: class for simple visualizations of monolayer systems

-CPovSTMSimpleBilayer: class for simple visualizations of bilayer systems
"""
from povray.ipovfromsource import IPovFromSource
from povray.predefined_povrayobjects.ccamera import CCameraForSpinLattice
from povray.utilities.spinconfiguration_utilities import calculate_phi_theta, calculate_colorcode_azimut
from povray.utilities.cfilterlattice import CFilterHexAroundCenter
from pathlib import Path
from typing import TypeVar, Tuple, List, Union
import vapory as vap
import numpy as np
import pandas as pd

PathLike = TypeVar("PathLike", str, Path)


class CPovSTMSimpleMono(IPovFromSource):
    r"""
    Class for simple visualization of STM source file.
    """

    def __call__(self, outputpath: PathLike = 'stmpov.png', width: float = 2000, height: float = 2000,
                 antialiasing: bool = False, antialiasingvalue: float = 0.001) -> None:
        r"""
        calls the rendering

        Args:
            outputpath(PathLike): name and path of created output. Default is stmpov in current directory
            width(float): width of the rendered image
            height(float): height of the rendered image
        """
        scene = vap.Scene(self._camera,
                          objects=self._objects,
                          included=['colors.inc', 'shapes.inc'])
        if antialiasing:
            scene.render(outfile=outputpath, width=width, height=height, antialiasing=antialiasingvalue)
        else:
            scene.render(outfile=outputpath, width=width, height=height)

    def __init__(self, sourcefile: PathLike, location: Union[None, List[float]] = None,
                 lookat: Union[None, List[float]] = None, viewangle: float = 90,
                 viewdistance: Union[float, None] = None, stretch: float = 1.5) -> None:
        r"""
        Initializes the visualization

        Args:
            sourcefile(PathLike): Path to source file of spin configuration

            location(List[float]): location of the observer. Default value is None. In that case the location will be
            automatically set with the use of the positions of the spins in the source file.

            lookat(List[float]): location of the observed point. Default value is None. In that case the location will
            be automatically set with the use of the positions of the spins in the source file.

            viewangle(float): Angle for vector from location to lookat. Default angle is 90 degree, which corresponds
            to a view from the top of the spin configuration. Will be used only if location and lookat are None

            viewdistance(float): Distance from the camera to the observed point. Default distance is None. Will be used
            only if location and lookat are None and if it is different from None. In case that location and lookat are
            not defined, a viewangle is selected and this parameter is None the distance will be calculated automatic
            from the spin configuration itself.

            stretch(float): Stretches the whole configuration by a factor for better visibility of cones

        """
        self._sourcefile = sourcefile
        self._position, self._spin = self._load_configuration(sourcefile=sourcefile)
        self._phi, self._theta, *_ = calculate_phi_theta(spin=self._spin)
        self._colorcode = calculate_colorcode_azimut(theta=self._theta)
        # stretching of lattice should be done before determining camera position. Otherwise camera could be wrong.
        self._position = self._position * stretch
        self._camera = CCameraForSpinLattice(self._position, location, lookat, viewangle, viewdistance).cam
        self._objects = []
        self._objects.append(self._load_background())
        self._objects.extend(self._load_lightsources())
        self._objects.extend(self._load_cones())

    @classmethod
    def _load_background(cls) -> vap.Background:
        r"""
        Returns:
            the background
        """
        return vap.Background('color Black')

    @classmethod
    def _load_lightsources(cls) -> List[vap.LightSource]:
        r"""
        Returns:
            the lightsources
        """
        return [vap.LightSource([12, -12, -12], 'color White', 'shadowless'),
                vap.LightSource([12, 0, -12], 'color White', 'area_light', [5, 0, 0], [0, 5, 0], 5, 5, 'adaptive', 1,
                                'jitter')]

    def _load_cones(self, length: float = 1.2, baseradius: float = 0.4) \
            -> List[vap.Cone]:
        r"""
        Create povray cones and returns them as a list

        Returns:
            A list of vapory cones which will be translated in povray cones when rendering
        """
        l_cones = []
        for (idx, pos) in enumerate(self._position):
            l_cones.append(vap.Cone([0.0, 0.0, length / 2], 0.0, [0.0, 0.0, -1 * length / 2], baseradius,
                                    'rotate', [0.0, self._theta[idx], 0.0],
                                    'rotate', [0.0, 0.0, self._phi[idx]],
                                    'translate', list(pos),
                                    vap.Texture(vap.Pigment('color', self._colorcode[idx]))))
        return l_cones

    @classmethod
    def _load_configuration(cls, sourcefile: PathLike) -> Tuple[np.ndarray, np.ndarray]:
        r"""
        Loads spin configuration from sourcefile. The source file is expected to have white spaces as separators and
        has to have the following structure in each line: x y z sx sy sz m (Usual structure of STM output from SpinD
        code)

        Args:
            sourcefile(PathLike): path to source file

        Returns:
            the positions, the spins
        """
        l_df = pd.read_csv(sourcefile, sep=r'\s+', header=None)
        l_position = l_df.iloc[:, :3].to_numpy()
        l_spin = l_df.iloc[:, 3:6].to_numpy()
        return l_position, l_spin

    def __repr__(self) -> str:
        r"""
        Returns:
            Representation of visualization class
        """
        return f"Representation of monolayer spin configuration with {len(self._position)} spins."

    @property
    def sourcefile(self) -> PathLike:
        r"""
        Returns:
             the source file
        """
        return self._sourcefile


class CPovSTMSimpleBilayer(IPovFromSource):
    r"""
    Class for simple visualization of STM source file.
    """

    def __call__(self, outputpath: PathLike = 'stmpov.png', width: float = 2000, height: float = 2000,
                 antialiasing: bool = False, antialiasingvalue: float = 0.001) -> None:
        r"""
        calls the rendering

        Args:
            outputpath(PathLike): name and path of created output. Default is stmpov in current directory
            width(float): width of the rendered image
            height(float): height of the rendered image
        """
        scene = vap.Scene(self._camera,
                          objects=self._objects,
                          included=['colors.inc', 'shapes.inc'])
        if antialiasing:
            scene.render(outfile=outputpath, width=width, height=height, antialiasing=antialiasingvalue)
        else:
            scene.render(outfile=outputpath, width=width, height=height)

    def __init__(self, sourcefile: PathLike, location: Union[None, List[float]] = None,
                 lookat: Union[None, List[float]] = None, viewangle: float = 0,
                 viewdistance: Union[float, None] = None, stretch: float = 1.51, filterlatt: bool = False,
                 filterkeyword: str = 'hexaroundcenter', percentage: float = 0.5) -> None:
        r"""
        Initializes the visualization

        Args:
            sourcefile(PathLike): Path to source file of spin configuration

            location(List[float]): location of the observer. Default value is None. In that case the location will be
            automatically set with the use of the positions of the spins in the source file.

            lookat(List[float]): location of the observed point. Default value is None. In that case the location will
            be automatically set with the use of the positions of the spins in the source file.

            viewangle(float): Angle for vector from location to lookat. Default angle is 90 degree, which corresponds
            to a view from the top of the spin configuration. Will be used only if location and lookat are None

            viewdistance(float): Distance from the camera to the observed point. Default distance is None. Will be used
            only if location and lookat are None and if it is different from None. In case that location and lookat are
            not defined, a viewangle is selected and this parameter is None the distance will be calculated automatic
            from the spin configuration itself.

            stretch(float): Stretches the whole configuration by a factor for better visibility of cones

            filterlatt(bool): boolean whether to filter the lattice to a specific region

            filterkeyword(str): decider string for factory patter, which filter algorithm shall be performed

            percentage(float): factor for some filtering algorithm. See the belonging implementation for details.
        """
        self._sourcefile = sourcefile
        self._position, self._spin = self._load_configuration(sourcefile=sourcefile)
        if filterlatt:
            self._filter_configuration(filterkeyword, percentage)
        self._phi, self._theta, *_ = calculate_phi_theta(spin=self._spin)
        self._colorcode = calculate_colorcode_azimut(theta=self._theta)
        self._position = self._position * stretch
        self._camera = CCameraForSpinLattice(self._position, location, lookat, viewangle, viewdistance,
                                             translatelookatz=5 * stretch).cam
        self._objects = []
        self._objects.append(self._load_background())
        self._objects.extend(self._load_lightsources())
        self._objects.extend(self._load_cones())

    def _filter_configuration(self, filterkeyword: str, percentage: float) -> None:
        r"""
        Sometimes one does not want to display the whole system. In most cases one only wants to look at an area around
        the center. This filtering operation takes the information for position and applies some filtering condition
        within the plane perpendicular to the z-direction. This is applied to all layers. The spin array is filtered the
        same way.

        Args:
            filterkeyword(str): Decider string for factory pattern of filter class.

            percentage(float): Some filters may need a percentage value to say how much of the lattice shall be filtered
            based on the original size.
        """
        if filterkeyword == 'hexaroundcenter':
            l_filter = CFilterHexAroundCenter(performfilteron=self._position, percentage=percentage)
        else:
            raise NotImplementedError('Filter not implemented yet.')
        self._position = l_filter.applyfilter(tofilter=self._position)
        self._spin = l_filter.applyfilter(tofilter=self._spin)

    @classmethod
    def _load_background(cls) -> vap.Background:
        r"""
        Returns:
            the background
        """
        return vap.Background('color Black')

    @classmethod
    def _load_lightsources(cls) -> List[vap.LightSource]:
        r"""
        Returns:
            the lightsources
        """
        return [vap.LightSource([12, -12, -12], 'color White', 'shadowless'),
                vap.LightSource([12, 0, -12], 'color White', 'area_light', [5, 0, 0], [0, 5, 0], 5, 5, 'adaptive', 1,
                                'jitter')]

    def _load_cones(self, length: float = 1.2, baseradius: float = 0.4) -> List[vap.Cone]:
        r"""
        Create povray cones and returns them as a list

        Returns:
            A list of vapory cones which will be translated in povray cones when rendering
        """
        l_cones = []
        for (idx, pos) in enumerate(self._position):
            l_cones.append(vap.Cone([0.0, 0.0, length / 2], 0.0, [0.0, 0.0, -1 * length / 2], baseradius,
                                    'rotate', [0.0, self._theta[idx], 0.0],
                                    'rotate', [0.0, 0.0, self._phi[idx]],
                                    'translate', [pos[0], pos[1], 10 * pos[2]],
                                    vap.Texture(vap.Pigment('color', self._colorcode[idx]))))
        return l_cones

    @classmethod
    def _load_configuration(cls, sourcefile: PathLike) -> Tuple[np.ndarray, np.ndarray]:
        r"""
        Loads spin configuration from sourcefile. The source file is expected to have white spaces as separators and
        has to have the following structure in each line: x y z sx sy sz m (Usual structure of STM output from SpinD
        code)

        Args:
            sourcefile(PathLike): path to source file

        Returns:
            the positions, the spins
        """
        l_df = pd.read_csv(sourcefile, sep=r'\s+', header=None)
        l_position = l_df.iloc[:, :3].to_numpy()
        l_spin = l_df.iloc[:, 3:6].to_numpy()
        return l_position, l_spin

    def __repr__(self) -> str:
        r"""
        Returns:
            Representation of visualization class
        """
        return f"Representation of bilayer spin configuration with {len(self._position)} spins."

    @property
    def sourcefile(self) -> PathLike:
        r"""
        Returns:
             the source file
        """
        return self._sourcefile



# -*- coding: utf-8 -*-
r"""
procedural rendering of spin configurations. An object oriented version is planned. There are methods for rendering
the povray source files generated from the SpinD-Kiel-code.
"""
from povray.cstmfromsource import CPovSTMSimpleBilayer
from povray.cvectorfromsource import CPovVectorSimpleBilayer
import numpy as np
edgelength= 10
shift = np.array([24.5*1.5,0.0])
vertices = np.array([[0.0, 0.0],[5,np.sqrt(3)/2*10],[10,0],[5,-1*np.sqrt(3)/2*10]])+shift

#CPovSTMSimpleBilayer(sourcefile='spin_0001.dat',filterlatt=True,percentage=0.3, viewdistance=35)()

CPovVectorSimpleBilayer(sourcefile='vector_sp.dat_0004.dat',viewangle=0, viewdistance=35, circleshift=np.array([0.0,0.0, 0.0]),polygonshift=np.array([1.5, np.sqrt(3) /4]), filterlatt=True,percentage=0.3, layerdistancefactor=5)()



# -*- coding: utf-8 -*-
r"""
Constants for the shellcommands library
"""
# SpinD path
SPIND_WINDOWS = r'C:\Users\Hendrik\source\repos\SpinD_Kiel\spin.exe'
SPIND_LINUX = r'/home/schrautzer/code/SpinD_Kiel/spin'


# -*- coding: utf-8 -*-
r"""
Classes for visualization of eigenvector configurations from STM files.

-CPovVectorSimpleMono: @todo

-CPovVectorSimpleBilayer: class for simple visualizations of bilayer systems
"""
from povray.ipovfromsource import IPovFromSource, PathLike
from povray.utilities.spinconfiguration_utilities import calculate_phi_theta, calculate_colorcode_azimut, \
    identifylayers, tabledist, determine_layer_centers
from povray.predefined_povrayobjects.ccamera import CCameraForSpinLattice
from povray.utilities.cfilterlattice import CFilterHexAroundCenter
import povray.utilities.colors as col
import vapory as vap
from typing import Union, List, Tuple
import pandas as pd
import numpy as np


class CPovVectorSimpleMono(IPovFromSource):
    r"""
    simple visualization class for eigenvectors for monolayer spin configurations.
    """

    def __init__(self) -> None:
        r"""
        initializes visualization of eigenvector for monolayer configuration
        """
        raise NotImplementedError('Visualizations of monolayer eigenvectors are not yet coded.')

    def __repr__(self) -> str:
        r"""
        Returns:
            representation of class
        """
        pass

    def __call__(self, outputpath: PathLike, width: float = 2000, height: float = 2000, antialiasing: bool = False,
                 antialiasingvalue: float = 0.001) -> None:
        r"""
        calls the rendering

        Args:
            outputpath(PathLike): name and path of created output. Default is stmpov in current directory
            width(float): width of the rendered image
            height(float): height of the rendered image
        """
        pass


class CPovVectorSimpleBilayer(IPovFromSource):
    r"""
    simple visualization class for eigenvectors for bilayer spin configurations.
    """

    def __init__(self, sourcefile: PathLike, location: Union[None, List[float]] = None,
                 lookat: Union[None, List[float]] = None, viewangle: float = 0,
                 viewdistance: Union[float, None] = None, stretch: float = 1.5, drawcircles: bool = True,
                 circleshift: np.ndarray = np.array([0.0, 0.0, 0.0]),filterlatt: bool = False,
                 filterkeyword: str = 'hexaroundcenter', percentage: float = 0.5, polygonshift: np.ndarray = np.array([0.0,0.0]), layerdistancefactor: float = 10) -> None:
        r"""
        Initializes the visualization

        Args:
            sourcefile(PathLike): Path to source file of spin configuration

            location(List[float]): location of the observer. Default value is None. In that case the location will be
            automatically set with the use of the positions of the spins in the source file.

            lookat(List[float]): location of the observed point. Default value is None. In that case the location will
            be automatically set with the use of the positions of the spins in the source file.

            viewangle(float): Angle for vector from location to lookat. Default angle is 90 degree, which corresponds
            to a view from the top of the spin configuration. Will be used only if location and lookat are None

            viewdistance(float): Distance from the camera to the observed point. Default distance is None. Will be used
            only if location and lookat are None and if it is different from None. In case that location and lookat are
            not defined, a viewangle is selected and this parameter is None the distance will be calculated automatic
            from the spin configuration itself.

            stretch(float): Stretches the whole configuration by a factor for better visibility of cones

            drawcircles(bool): For better three dimensional visibility the next neighbors can be elucidated via circles.

            filterlatt(bool): boolean whether to filter the lattice to a specific region

            filterkeyword(str): decider string for factory patter, which filter algorithm shall be performed

            percentage(float): factor for some filtering algorithm. See the belonging implementation for details.

            layerdistancefactor(float): factor for the distance between the layers within the povray rendering

            polygonshift(np.ndarray): whether the polygon is shifted. This is necessary if the center of the filtering
            shouldn't be in the center of the lattice
        """
        self._sourcefile = sourcefile
        self._position, self._eigenvector = self._load_configuration(sourcefile=sourcefile)
        if filterlatt:
            self._filter_configuration(filterkeyword, percentage, polygonshift)
        # stretch configuration by factor
        self._position = self._position * stretch
        self._phi, self._theta, self._length = calculate_phi_theta(spin=self._eigenvector)
        self._colorcode = calculate_colorcode_azimut(theta=self._theta)
        self._stretch = stretch
        self._camera = CCameraForSpinLattice(self._position, location, lookat, viewangle, viewdistance,
                                             translatelookatz=5 * stretch).cam
        self._objects = []
        self._objects.append(self._load_background())
        self._objects.extend(self._load_lightsources())
        self._objects.extend(self._load_vectors(layerdistancefactor=layerdistancefactor))
        if drawcircles:
            self._objects.extend(self._load_circles(circleshift=circleshift, layerdistancefactor=layerdistancefactor))

    def _filter_configuration(self, filterkeyword: str, percentage: float, polygonshift: np.ndarray) -> None:
        r"""
        Sometimes one does not want to display the whole system. In most cases one only wants to look at an area around
        the center. This filtering operation takes the information for position and applies some filtering condition
        within the plane perpendicular to the z-direction. This is applied to all layers. The spin array is filtered the
        same way.

        Args:
            filterkeyword(str): Decider string for factory pattern of filter class.

            percentage(float): Some filters may need a percentage value to say how much of the lattice shall be filtered
            based on the original size.

            polygonshift(np.ndarray): whether the polygon is shifted. This is necessary if the center of the filtering
            shouldn't be in the center of the lattice
        """
        if filterkeyword == 'hexaroundcenter':
            l_filter = CFilterHexAroundCenter(performfilteron=self._position,polygonshift=polygonshift, percentage=percentage)
        else:
            raise NotImplementedError('Filter not implemented yet.')
        self._position = l_filter.applyfilter(tofilter=self._position)
        self._eigenvector = l_filter.applyfilter(tofilter=self._eigenvector)

    def _load_circles(self, circleshift: np.ndarray,layerdistancefactor:float, numberofcircles: int = 4, circlewidth: float = 0.025) -> List[
        vap.Object]:
        r"""
        Supports the visibility of the eigenvector representation via next neighbor shells. Written for arbitrary
        lattice types and arbitray numbers of layers. The reference point is the centroid of the layer.

        Args:
            circleshift(np.ndarray): Shift vector for all circles. This is sometimes necessary as the center of the
            circle is determined by the center of the lattice but the skyrmion or saddlepoint isnt alway directly
            centered (3-spin center)

            layerdistancefactor(float): factor for the difference between the layers.

            numberofcircles(int): if the parameter "drawcircles" is enabled this parameter defines the number of next
            neighbor shells drawn.

            circlewidth(float): width of the circle border
        """
        circles = []
        l_n, l_zcoords = identifylayers(self._position)
        for n in range(l_n):
            l_pos_layer = self._position[np.abs(self._position[:, 2] - l_zcoords[n]) <= 1e-3]
            l_centroidoflayer = np.array([coord if coord >=1e-8 else 0 for coord in np.mean(l_pos_layer, axis=0)])
            l_centroidoflayer[2] = 0.0
            l_shells = tabledist(l_pos_layer[:, :2], referencepoint=l_centroidoflayer[:2] + circleshift[:2])[:numberofcircles]
            l_circlecenter = l_centroidoflayer + np.array([0.0, 0.0, layerdistancefactor * l_zcoords[n]]) + circleshift
            for shell in l_shells:
                circles.append(vap.Object(vap.Macro('Circle_Line', l_circlecenter, shell, circlewidth, [0.0, 0.0, 1.0]),
                                          vap.Texture(vap.Pigment('color', col.lightgray()))))
        return circles

    def _load_vectors(self,layerdistancefactor: float, headlength: float = 0.35, headwidth: float = 0.15, tailwidth: float = 0.08,
                      scale: float = 10, lowerlengthlimit: float = 1e-5) -> List[vap.Union]:
        r"""
        Create unions of povray cones and cylinders which represent arrows and returns them as a list. The union of the
        cone and the cylinder corresponds to the length L of the certain vector. The position of each vector is meant to
        be L / 2.

        If the length of the whole arrow is smaller than the headlength only the cone is displayed. This means that
        really short vectors are not displayed with the correct length.

        Args:
            layerdistancefactor(float): factor for the difference between the layers.

            headlength(float): height of the cone representing the head of the arrow (in real length)

            headwidth(float): base radius of the cone representing the head of the arrow

            tailwidth(float): radius of the cylinder representing the tail of the arrow

            scale(float): generally eigenvectors are small. For a visibility the scale parameter multiplies the length
            of each vector by this fixed "scale" value.

            lowerlengthlimit(float): Vectors which are shorter than this parameter are displayed as spheres.

        Returns:
            A list of vapory objects which will be translated in povray objects when rendering
        """
        l_vectors = []
        for (idx, pos) in enumerate(self._position):
            if self._length[idx] >= lowerlengthlimit:
                ll = scale * self._length[idx]
                if ll < headlength:
                    l_vectors.append(vap.Cone([0.0, 0.0, -1 * headlength / 2], headwidth,
                                     [0.0, 0.0, headlength / 2], 0,
                              'rotate', [0.0, self._theta[idx], 0.0],
                              'rotate', [0.0, 0.0, self._phi[idx]],
                              'translate', [pos[0], pos[1], layerdistancefactor * pos[2]],
                              vap.Texture(vap.Pigment('color', self._colorcode[idx]))))
                else:
                    l_cylbase = [0.0, 0.0, -1 * ll / 2]
                    l_cylcap = [0.0, 0.0, ll / 2 - headlength]
                    l_conecap = [0.0, 0.0, ll / 2]
                    l_vectors.append(
                        vap.Union(vap.Cylinder(l_cylbase, l_cylcap, tailwidth),
                                  vap.Cone(l_cylcap, headwidth, l_conecap, 0),
                                  'rotate', [0.0, self._theta[idx], 0.0],
                                  'rotate', [0.0, 0.0, self._phi[idx]],
                                  'translate', [pos[0], pos[1], layerdistancefactor * pos[2]],
                                  vap.Texture(vap.Pigment('color', self._colorcode[idx]))))
            else:
                l_vectors.append(
                    vap.Sphere([pos[0], pos[1], layerdistancefactor * pos[2]], tailwidth,
                               vap.Texture(vap.Pigment('color', self._colorcode[idx])))
                )
        return l_vectors

    @classmethod
    def _load_background(cls) -> vap.Background:
        r"""
        Returns:
            the background
        """
        return vap.Background('color Black')

    @classmethod
    def _load_lightsources(cls) -> List[vap.LightSource]:
        r"""
        Returns:
            the lightsources
        """
        return [vap.LightSource([12, -12, -12], 'color White', 'shadowless'),
                vap.LightSource([12, 0, -12], 'color White', 'area_light', [5, 0, 0], [0, 5, 0], 5, 5, 'adaptive', 1,
                                'jitter')]

    @classmethod
    def _load_configuration(cls, sourcefile: PathLike) -> Tuple[np.ndarray, np.ndarray]:
        r"""
        Loads eigenvector of spin configuration from sourcefile. The source file is expected to have white spaces as
        separators and has to have the following structure in each line: x y z vx vy vz v (Usual structure of
        eigenvector output from SpinD code)

        Args:
            sourcefile(PathLike): path to source file

        Returns:
            the positions, the vectors
        """
        l_df = pd.read_csv(sourcefile, sep=r'\s+', header=None)
        l_position = l_df.iloc[:, :3].to_numpy()
        l_vector = l_df.iloc[:, 3:6].to_numpy()
        return l_position, l_vector

    def __repr__(self) -> str:
        pass

    def __call__(self, outputpath: PathLike = 'vecpov.png', width: float = 2000, height: float = 2000,
                 antialiasing: bool = False, antialiasingvalue: float = 0.001) -> None:
        r"""
        calls the rendering

        Args:
            outputpath(PathLike): name and path of created output. Default is stmpov in current directory
            width(float): width of the rendered image
            height(float): height of the rendered image
        """
        scene = vap.Scene(self._camera,
                          objects=self._objects,
                          included=['colors.inc', 'shapes.inc', 'analytical_g.inc'])
        if antialiasing:
            scene.render(outfile=outputpath, width=width, height=height, antialiasing=antialiasingvalue)
        else:
            scene.render(outfile=outputpath, width=width, height=height)



# -*- coding: utf-8 -*-
r"""
Gets and summarizes the results of a minimization executed with spindyanmic of the spinD code for a system with a
skyrmion.
"""
import pandas as pd
import numpy as np
from results.igetresults import IResults
from pathlib import Path
from typing import Union, List, Tuple, Dict
from python3.magnetisations import SpinLattice, MultiLayer


class CSDSkyrmion(IResults):
    r"""
    This class can be called in folder after a minimization of a skyrmionic system with the SpinD code. This system can
    be a monolayer or a multilayer.
    """

    def __init__(self, calc_dir: Path = Path.cwd(), nlayer: int = 1, signaturename: str = 'signature',
                 signaturevalue: Union[float, str] = 0) -> None:
        r"""
        Initializes the result-summarizing process.

        Args:
            calc_dir(Path): directory of the spinD calculation. Default is the current working directory.

            nlayer(int): number of layers for the system.

            signaturename(str): if multiple calculations are evaluated with the goal to connect the dataframes then you
            can define a signaturename which the column names which describes what quantity changes from calc. to calc.
            It can be for example magnetic field or interlayer exchange

            signaturevalue(float, str): The value for the current calculation
        """
        self.calc_dir = calc_dir
        self.nlayer = nlayer
        self.signaturename = signaturename
        self.signaturevalue = signaturevalue
        self.i_en_dens = (calc_dir / 'energy_density_end.dat').is_file()
        self.i_conv = (calc_dir / 'convergence.dat').is_file()
        self.i_job = (calc_dir / 'job.out').is_file()
        self.i_stmend = (calc_dir / 'SpinSTM_end.dat').is_file()
        self.i_stmi = (calc_dir / 'SpinSTMi.dat').is_file()

        if not any([self.i_en_dens, self.i_conv, self.i_job, self.i_stmend, self.i_stmi]):
            print('None of the following files is present: energy_density_end.dat, convergence.dat, job.out')
            print('SpinSTM_end.dat, SpinSTMi.dat -> finished.')

    def __call__(self) -> pd.DataFrame:
        r"""
        Returns:
            the DataFrame with the information from the spindynamics calculation
        """
        laststep, totalenergy, torque = self.convergencefile()
        skprofile_i, skprofile_f = self.skyrmions()
        densities = self.energydensity()

        d = {self.signaturename: self.signaturevalue,
             't_computation': self.computationtime(),
             'steps': laststep,
             'en_per_UC': totalenergy,
             'torque': torque}
        d.update(skprofile_i)
        d.update(skprofile_f)
        d.update(densities)
        return pd.DataFrame(d, index=[self.signaturevalue])

    def energydensity(self) -> Dict[str, Union[None,float]]:
        r"""
        Reads the energy_density_end.dat file if the energy density file exists

        Returns:
            a dictionary containing the sum of the nergy contributions for each layer
        """
        if not self.i_en_dens:
            endict = {}
            for idx in range(self.nlayer):
                endict[f'ex_intra_lay{idx}'] = None
                endict[f'ex_inter_lay{idx}'] = None
                endict[f'dmi_lay{idx}'] = None
                endict[f'ani_lay{idx}'] = None
                endict[f'zee_lay{idx}'] = None
                endict[f'dip_lay{idx}'] = None
                endict[f'biq_lay{idx}'] = None
                endict[f'four_lay{idx}'] = None
                endict[f'tot_lay{idx}'] = None
            return endict
        df = pd.read_csv(self.calc_dir / 'energy_density_end.dat', sep=r'\s+', index_col=False)
        df.columns = ['x', 'y', 'z', 'ex_intra', 'ex_inter', 'dmi', 'ani', 'zee', 'dip', 'biq', 'four', 'tot']
        z_layer = df.z.unique()
        endict = {}
        for (idx, zl) in enumerate(z_layer):
            dfl = df[df['z'] == zl]
            endict[f'ex_intra_lay{idx}'] = dfl.ex_intra.sum()
            endict[f'ex_inter_lay{idx}'] = dfl.ex_inter.sum()
            endict[f'dmi_lay{idx}'] = dfl.dmi.sum()
            endict[f'ani_lay{idx}'] = dfl.ani.sum()
            endict[f'zee_lay{idx}'] = dfl.zee.sum()
            endict[f'dip_lay{idx}'] = dfl.dip.sum()
            endict[f'biq_lay{idx}'] = dfl.biq.sum()
            endict[f'four_lay{idx}'] = dfl.four.sum()
            endict[f'tot_lay{idx}'] = dfl.tot.sum()
        return endict

    def skyrmions(self) -> Tuple[Dict[str, Union[float, None]], Dict[str, Union[float, None]]]:
        r"""
        Calculates the skyrmion profiles for each layer for the SpinSTMi file and the SpinSTM_end file.
        """
        if self.i_stmi:
            stmiresult = self.skyrmionprofile(self.calc_dir / 'SpinSTMi.dat')
        else:
            stmiresult = {}
            for n in range(self.nlayer):
                stmiresult[f'rad_lay{n}'] = None
                stmiresult[f'R0x_lay{n}'] = None
                stmiresult[f'R0y_lay{n}'] = None
                stmiresult[f'c_lay{n}'] = None
                stmiresult[f'w_lay{n}'] = None
        if self.i_stmend:
            stmendresult = self.skyrmionprofile(self.calc_dir / 'SpinSTM_end.dat')
        else:
            stmendresult = {}
            for n in range(self.nlayer):
                stmendresult[f'rad_lay{n}'] = None
                stmendresult[f'R0x_lay{n}'] = None
                stmendresult[f'R0y_lay{n}'] = None
                stmendresult[f'c_lay{n}'] = None
                stmendresult[f'w_lay{n}'] = None
        return stmiresult, stmendresult

    def skyrmionprofile(self, file: Path) -> Dict[str, float]:
        r"""
        Reads the spin lattice from a SpinSTM file, evaluates the skyr. profile with the use of the userlib-spinD-lib.

        Args:
            file(Path): SpinSTM-file

        Returns:
            The dict with the following keys:
            skj_rad : radius of sk in the j - layer
            skj_R0x : x-component of the center of sk in the j - layer
            skj_R0y : y-component of the center of sk in the j - layer
            skj_c : c-param of the profile of sk in the j - layer
            skj_w : c-param of the profile of sk in the j - layer
        """
        ML = MultiLayer(path=str(file), number_layers=self.nlayer)
        for lay in ML.Layer:
            lay.getSkradius()
            profiles = {}
        for n in range(self.nlayer):
            if ML.Layer[n].sk_radius is None:
                profiles[f'rad_lay{n}'] = None
                profiles[f'R0x_lay{n}'] = None
                profiles[f'R0y_lay{n}'] = None
                profiles[f'c_lay{n}'] = None
                profiles[f'w_lay{n}'] = None
            else:
                profiles[f'rad_lay{n}'] = ML.Layer[n].sk_radius
                profiles[f'R0x_lay{n}'] = ML.Layer[n].sk_popt[0]
                profiles[f'R0y_lay{n}'] = ML.Layer[n].sk_popt[1]
                profiles[f'c_lay{n}'] = ML.Layer[n].sk_popt[2]
                profiles[f'w_lay{n}'] = ML.Layer[n].sk_popt[3]
        return profiles

    def convergencefile(self) -> Union[Tuple[float, float, float], Tuple[None, None, None]]:
        r"""
        Reads the convergence file of the calculation if it exists.

        Returns:
            the number of steps necessary for convergence, the total energy and the last torque. If the file doesn't
            exist it returns None for all three of them
        """
        if not self.i_conv:
            return None, None, None
        for line in reversed(open(str(self.calc_dir / 'convergence.dat')).readlines()):
            L = line.split()
            break
        return int(L[0]), float(L[1]), float(L[2])

    def computationtime(self) -> Union[float, None]:
        r"""
        Checks the computation time of the calculation

        Returns:
            either the computation time written in job.out or None if job.out doesn't exist or the specific line is
            not present in job.out
        """
        if not self.i_job:
            return None
        line_exi = False
        with open(str(self.calc_dir / 'job.out')) as f:
            for line in f:
                if line.lstrip().startswith('computation time:'):
                    line_exi = True
                    return float(line.split()[2])
        if not line_exi:
            return None
